from collections import defaultdict

from jaxoplanet.starry.core.basis import U
from jaxoplanet.starry.core.polynomials import Pijk
from jaxoplanet.starry.multiprecision import mp, utils
from jaxoplanet.starry.multiprecision.basis import A1, A2
from jaxoplanet.starry.multiprecision.rotation import (
    dot_rotation_matrix,
    dot_rz,
    get_R,
)
from jaxoplanet.starry.multiprecision.solution import get_sT, rT

CACHED_MATRICES = defaultdict(
    lambda: {
        "R_obl": {},
        "R_inc": {},
        "sT": {},
    }
)


def flux(ydeg=0, udeg=0, inc=mp.pi / 2, obl=0.0, cache=None):
    if cache is None:
        cache = CACHED_MATRICES

    deg = ydeg + udeg

    _rT = rT(deg)
    _A2 = A2(deg, cache=cache)

    if ydeg > 0:
        _A1 = A1(ydeg, cache=cache)
        R_px = get_R("R_px", ydeg, cache=cache)
        R_mx = get_R("R_mx", ydeg, cache=cache)
        R_obl = get_R("R_obl", ydeg, obl=obl, inc=inc, cache=cache)
        R_inc = get_R("R_inc", ydeg, obl=obl, inc=inc, cache=cache)
    else:
        _A1 = A1(0, cache=cache)

    if udeg > 0:
        rT_udeg = rT(udeg)

    def rotate_y(y, phi):
        y_rotated = dot_rotation_matrix(ydeg, None, None, R_px)(y)
        y_rotated = dot_rz(ydeg, phi)(y_rotated)
        y_rotated = dot_rotation_matrix(ydeg, None, None, R_mx)(y_rotated)
        y_rotated = dot_rotation_matrix(ydeg, None, None, R_obl)(y_rotated)
        y_rotated = dot_rotation_matrix(ydeg, None, None, R_inc)(y_rotated)
        return y_rotated

    def rot_flux(y, phi):
        y = mp.matrix(y.tolist())
        y_rotated = rotate_y(y, phi)
        return ((_A1 @ y_rotated).T @ _rT)[0]

    def occ_flux(y, b, r, phi=0.0, u=None, rotate=True):
        xo = 0.0
        yo = b
        theta_z = mp.atan2(xo, yo)
        _sT = get_sT(deg, b, r, cache=cache)
        x = _sT.T @ _A2

        if ydeg > 0:
            y = mp.matrix(y.tolist())
            if rotate:
                y_rotated = rotate_y(y, phi)
                y_rotated = dot_rz(ydeg, theta_z)(y_rotated)
        else:
            y_rotated = mp.matrix([1])

        py = (_A1 @ y_rotated).T

        if udeg > 0:
            _U = utils.to_mp(U(udeg))
            u = mp.matrix([1, *u])
            pu = u.T @ _U
            norm = mp.pi / (pu @ rT_udeg)[0]
            py = Pijk.from_dense(pu.tolist()[0]) * Pijk.from_dense(py)
            py = mp.matrix(
                [py.data.get(Pijk.n2ijk(i), 0.0) for i in range((py.degree + 1) ** 2)]
            ).T
        else:
            norm = 1
        return norm * py @ x.T

    def impl(b, r, y=(), phi=0.0, u=(), rotate=True):
        assert len(u) == udeg, "u must have length udeg"
        if ydeg > 0:
            assert len(y) == (ydeg + 1) ** 2, "y must have length (ydeg + 1) ** 2"

        if abs(b) >= (r + 1):
            return rot_flux(y, phi)
        elif r > 1 and abs(b) <= (r - 1):
            return mp.mpf(0.0)
        else:
            return occ_flux(y, b, r, phi, u=u, rotate=rotate)[0]

    return impl



__all__ = ["light_curve"]

from collections.abc import Callable
from functools import partial

import jax.numpy as jnp
import jpu.numpy as jnpu

from jaxoplanet import units
from jaxoplanet.core.limb_dark import light_curve as _limb_dark_light_curve
from jaxoplanet.light_curves.utils import vectorize
from jaxoplanet.proto import LightCurveOrbit
from jaxoplanet.types import Array, Quantity
from jaxoplanet.units import unit_registry as ureg


def light_curve(
    orbit: LightCurveOrbit, *u: Array, order: int = 10
) -> Callable[[Quantity], Array]:
    """Compute the light curve for arbitrary polynomial limb darkening

    See `Agol et al. (2020) <https://arxiv.org/abs/1908.03222>`_ and
    :func:`jaxoplanet.core.limb_dark.light_curve` for more technical details.

    Args:
        orbit (LightCurveOrbit): An orbit object that can be used to evaluate the
            relative positions of the transiting body with respect to the light source.
        u (Array): The coefficients of the polynomial limb darkening
        order (int): The order of the numerical integration used by the backend; see
            :func:`jaxoplanet.core.limb_dark.light_curve`

    Returns:
        A function which takes the time in days as input and returns the light curve flux
    """

    if u:
        ld_u = jnp.concatenate([jnp.atleast_1d(jnp.asarray(u_)) for u_ in u], axis=0)
    else:
        ld_u = jnp.array([])

    @units.quantity_input(time=ureg.d)
    @vectorize
    def light_curve_impl(time: Quantity) -> Array:
        if jnpu.ndim(time) != 0:
            raise ValueError(
                "The time passed to 'light_curve' has shape "
                f"{jnpu.shape(time)}, but a scalar was expected; "
                "this shouldn't typically happen so please open an issue "
                "on GitHub demonstrating the problem"
            )

        # Evaluate the coordinates of the transiting body
        r_star = orbit.central_radius
        x, y, z = orbit.relative_position(time)

        b = jnpu.sqrt(x**2 + y**2) / r_star
        assert b.units == ureg.dimensionless
        r = orbit.radius / r_star
        assert r.units == ureg.dimensionless

        lc_func = partial(_limb_dark_light_curve, ld_u, order=order)
        lc = lc_func(b.magnitude, r.magnitude)
        lc = jnp.where(z > 0, lc, 0)

        return lc

    return light_curve_impl



# mypy: ignore-errors

import warnings

import jax
import numpy as np
import pytest

from jaxoplanet.starry.core.solution import kappas, solution_vector
from jaxoplanet.test_utils import assert_allclose


def test_kappas():
    def _phi(b, r):
        arg = 1 - r**2 - b**2
        m = (r > 0) & (b > 0)
        arg[m] /= 2 * r[m] * b[m]
        result = np.arcsin(np.clip(arg, -1, 1))
        result[~m] = 0.5 * np.pi * np.sign(arg[~m])
        return result

    def _lam(b, r):
        arg = 1 - r**2 + b**2
        m = (r > 0) & (b > 0)
        arg[m] /= 2 * b[m]
        result = np.arcsin(np.clip(arg, -1, 1))
        result[~m] = 0.5 * np.pi * np.sign(arg[~m])
        return result

    b, r = np.meshgrid(np.linspace(0, 2, 100), np.linspace(0.01, 1.5, 13))
    b = b.flatten()
    r = r.flatten()
    kappa0, kappa1 = jax.vmap(kappas)(b, r)
    phi = _phi(b, r)
    lam = _lam(b, r)

    assert_allclose(kappa0, phi + 0.5 * np.pi)
    assert_allclose(kappa1, 0.5 * np.pi - lam)


@pytest.mark.parametrize("r", [0.01, 0.1, 1.0, 10.0, 100.0])
def test_solution(r, l_max=5, order=500):
    pytest.importorskip("mpmath")
    from jaxoplanet.starry.multiprecision import (
        solution as mp_solution,
        utils,
    )

    # We know that these are were the errors are the highest
    b = 1 - r if r < 1 else r

    expect = utils.to_numpy(mp_solution.sT(l_max, b, r))
    calc = solution_vector(l_max, order=order)(b, r)

    for n in range(expect.size):
        # For logging/debugging purposes, work out the case id
        l = np.floor(np.sqrt(n)).astype(int)  # noqa
        m = n - l**2 - l
        mu = l - m
        nu = l + m
        if mu == 1 and l == 1:
            case = 1
        elif mu % 2 == 0 and (mu // 2) % 2 == 0:
            case = 2
        elif mu == 1 and l % 2 == 0:
            case = 3
        elif mu == 1:
            case = 4
        elif (mu - 1) % 2 == 0 and ((mu - 1) // 2) % 2 == 0:
            case = 5
        else:
            case = 0

        assert_allclose(
            calc[n],
            expect[n],
            err_msg=f"n={n}, l={l}, m={m}, mu={mu}, nu={nu}, case={case}",
            atol=1e-15,
        )


@pytest.mark.parametrize("r", [0.1, 1.1])
def test_solution_compare_starry(r, l_max=10, order=500):
    starry = pytest.importorskip("starry")
    theano = pytest.importorskip("theano")
    theano.config.gcc__cxxflags += " -fexceptions"

    b = np.linspace(0, 1 + r, 501)[:-1]
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        m = starry.Map(l_max)
        s_expect = m.ops.sT(b, r)
    s_calc = solution_vector(l_max, order=order)(b, r)

    for n in range(s_expect.shape[1]):
        # For logging/debugging purposes, work out the case id
        l = np.floor(np.sqrt(n)).astype(int)  # noqa
        m = n - l**2 - l
        mu = l - m
        nu = l + m
        if mu == 1 and l == 1:
            case = 1
        elif mu % 2 == 0 and (mu // 2) % 2 == 0:
            case = 2
        elif mu == 1 and l % 2 == 0:
            case = 3
        elif mu == 1:
            case = 4
        elif (mu - 1) % 2 == 0 and ((mu - 1) // 2) % 2 == 0:
            case = 5
        else:
            case = 0

        assert_allclose(
            s_calc[:, n],
            s_expect[:, n],
            err_msg=f"n={n}, l={l}, m={m}, mu={mu}, nu={nu}, case={case}",
            atol=1e-6,
        )


def test_r_greater_one_grad():
    # this is an even more minimal version of what caused issue #235
    assert np.isfinite(jax.jacrev(solution_vector(3))(0.5, 10.0)).all()



# mypy: ignore-errors

import nox

ALL_PYTHON_VS = ["3.10", "3.11", "3.12"]


@nox.session(python=ALL_PYTHON_VS)
def test(session):
    session.install(".[test,test-math]")
    session.run("pytest", "-n", "auto", *session.posargs)


@nox.session(python=ALL_PYTHON_VS)
def test_x64(session):
    session.install(".[test,test-math]")
    env = {"JAX_ENABLE_X64": "1"}
    session.run("pytest", "-n", "auto", *session.posargs, env=env)


@nox.session(python=["3.10"])
def comparison(session):
    session.install(".[test,comparison]", "numpy<1.22")
    session.run("python", "-c", "import starry")
    session.run("python", "-c", "import theano")
    if session.posargs:
        args = session.posargs
    else:
        args = ("tests/starry",)
    session.run(
        "pytest",
        "-n",
        "auto",
        *args,
        env={"JAX_ENABLE_X64": "True"},
    )


@nox.session
def docs(session):
    session.install(".[docs]")
    with session.chdir("docs"):
        session.run(
            "python",
            "-m",
            "sphinx",
            "-T",
            "-E",
            "-W",
            "--keep-going",
            "-b",
            "dirhtml",
            "-d",
            "_build/doctrees",
            "-D",
            "language=en",
            ".",
            "_build/dirhtml",
        )



import jax.numpy as jnp
import pytest

from jaxoplanet.light_curves import transforms
from jaxoplanet.test_utils import assert_allclose
from jaxoplanet.units import quantity_input, unit_registry as ureg


@quantity_input(time=ureg.day)
def lc_func1(time):
    return jnp.ones_like(time.magnitude)


@quantity_input(time=ureg.day)
def lc_func2(time):
    return jnp.stack([0.5 * time.magnitude + 0.1, -1.5 * time.magnitude + 3.6], axis=-1)


@pytest.mark.parametrize("order", [0, 1, 2])
@pytest.mark.parametrize("lc_func", [lc_func1, lc_func2])
def test_integrate_invariant(order, lc_func):
    time = jnp.linspace(0, 10, 50)
    int_func = transforms.integrate(lc_func, exposure_time=0.1, order=order)
    assert_allclose(int_func(time), lc_func(time))



"""This module provides the core functionality to solve Kepler's equation in JAX. For
more details, see the :ref:`core-from-scratch` tutorial.
"""

__all__ = ["kepler"]

import jax
import jax.numpy as jnp
from jax.interpreters import ad

from jaxoplanet.types import Array


@jax.jit
def kepler(M: Array, ecc: Array) -> tuple[Array, Array]:
    """Solve Kepler's equation to compute the true anomaly

    Args:
        M: Mean anomaly
        ecc: Eccentricity

    Returns:
        The sine and cosine of the true anomaly
    """
    return _kepler(M, ecc)


@jax.custom_jvp
def _kepler(M: Array, ecc: Array) -> tuple[Array, Array]:
    # Wrap into the right range
    M = M % (2 * jnp.pi)

    # We can restrict to the range [0, pi)
    high = M > jnp.pi
    M = jnp.where(high, 2 * jnp.pi - M, M)

    # Solve
    ome = 1 - ecc
    E = starter(M, ecc, ome)
    E = refine(M, ecc, ome, E)

    # Re-wrap back into the full range
    E = jnp.where(high, 2 * jnp.pi - E, E)

    # Convert to true anomaly; tan(0.5 * f)
    tan_half_f = jnp.sqrt((1 + ecc) / (1 - ecc)) * jnp.tan(0.5 * E)
    tan2_half_f = jnp.square(tan_half_f)

    # Then we compute sin(f) and cos(f) using:
    #  sin(f) = 2*tan(0.5*f)/(1 + tan(0.5*f)^2), and
    #  cos(f) = (1 - tan(0.5*f)^2)/(1 + tan(0.5*f)^2)
    denom = 1 / (1 + tan2_half_f)
    sinf = 2 * tan_half_f * denom
    cosf = (1 - tan2_half_f) * denom

    return sinf, cosf


@_kepler.defjvp
def _(primals, tangents):
    M, e = primals
    M_dot, e_dot = tangents
    sinf, cosf = _kepler(M, e)

    # Pre-compute some things
    ecosf = e * cosf
    ome2 = 1 - e**2

    def make_zero(tan):
        if type(tan) is ad.Zero:
            return ad.zeros_like_aval(tan.aval)
        else:
            return tan

    # Propagate the derivatives
    f_dot = make_zero(M_dot) * (1 + ecosf) ** 2 / ome2**1.5
    f_dot += make_zero(e_dot) * (2 + ecosf) * sinf / ome2

    return (sinf, cosf), (cosf * f_dot, -sinf * f_dot)


def starter(M: Array, ecc: Array, ome: Array) -> Array:
    M2 = jnp.square(M)
    alpha = 3 * jnp.pi / (jnp.pi - 6 / jnp.pi)
    alpha += 1.6 / (jnp.pi - 6 / jnp.pi) * (jnp.pi - M) / (1 + ecc)
    d = 3 * ome + alpha * ecc
    alphad = alpha * d
    r = (3 * alphad * (d - ome) + M2) * M
    q = 2 * alphad * ome - M2
    q2 = jnp.square(q)
    w = jnp.square(jnp.cbrt(jnp.abs(r) + jnp.sqrt(q2 * q + r * r)))
    return (2 * r * w / (jnp.square(w) + w * q + q2) + M) / d


def refine(M: Array, ecc: Array, ome: Array, E: Array) -> Array:
    sE = E - jnp.sin(E)
    cE = 1 - jnp.cos(E)

    f_0 = ecc * sE + E * ome - M
    f_1 = ecc * cE + ome
    f_2 = ecc * (E - sE)
    f_3 = 1 - f_1
    d_3 = -f_0 / (f_1 - 0.5 * f_0 * f_2 / f_1)
    d_4 = -f_0 / (f_1 + 0.5 * d_3 * f_2 + (d_3 * d_3) * f_3 / 6)
    d_42 = d_4 * d_4
    dE = -f_0 / (f_1 + 0.5 * d_4 * f_2 + d_4 * d_4 * f_3 / 6 - d_42 * d_4 * f_2 / 24)

    return E + dE



import jax.numpy as jnp
import pytest

from jaxoplanet.object_stack import ObjectStack
from jaxoplanet.test_utils import assert_quantity_pytree_allclose


def f1(x):
    return x["a"] ** 2


def f2(x):
    return {"y": x["a"] ** 2}


@pytest.mark.parametrize(
    "func, objects, result",
    [
        (
            f1,
            (
                {"a": 0.5},
                {"a": 0.1},
            ),
            jnp.asarray([0.5**2, 0.1**2]),
        ),
        (
            f1,
            (
                {"a": 0.5},
                {"a": 0.1, "b": 1000.0},
            ),
            jnp.asarray([0.5**2, 0.1**2]),
        ),
        (
            f2,
            (
                {"a": 0.5},
                {"a": 0.1},
            ),
            {"y": jnp.asarray([0.5**2, 0.1**2])},
        ),
        (
            f2,
            (
                {"a": 0.5},
                {"a": 0.1, "b": 1000.0},
            ),
            {"y": jnp.asarray([0.5**2, 0.1**2])},
        ),
    ],
)
def test_object_stack(func, objects, result):
    stack = ObjectStack(*objects)
    assert_quantity_pytree_allclose(stack.vmap(func)(), result)



from collections import defaultdict
from functools import reduce

import numpy as np

from jaxoplanet.starry.core import basis
from jaxoplanet.starry.multiprecision import mp, utils
from jaxoplanet.starry.multiprecision.utils import (
    fac as fac_function,
    kron_delta,
)

lmax = 20

FAC_CACHE = {}
CACHED_MATRICES = defaultdict(lambda: {})


def get_A(A, l_max, cache=None):
    if cache is None:
        cache = CACHED_MATRICES

    name = A.__name__.replace("_", "")
    if name not in cache[l_max]:
        print(f"pre-computing {name}...")
        cache[l_max][name] = A(l_max)
    return cache[l_max][name]


def fac(n):
    if n not in FAC_CACHE:
        FAC_CACHE[n] = fac_function(n)

    return FAC_CACHE[n]


def A(l, m):
    """A spherical harmonic normalization constant."""
    return mp.sqrt(
        (2 - kron_delta(m, 0)) * (2 * l + 1) * fac(l - m) / (4 * mp.pi * fac(l + m))
    )


def B(l, m, j, k):
    """Another spherical harmonic normalization constant."""

    ratio = fac((l + m + k - 1) / 2) / fac((-l + m + k - 1) / 2)

    return 2**l * fac(m) / (fac(j) * fac(k) * fac(m - j) * fac(l - m - k)) * ratio


def C(p, q, k):
    """Return the binomial theorem coefficient `C`."""
    return fac(k / 2) / (fac(q / 2) * fac((k - p) / 2) * fac((p - q) / 2))


def Y(l, m):
    """Return the spherical harmonic of degree `l` and order `m`."""
    coeffs = defaultdict(lambda: {})

    def get(function, *indices):
        fun_name = function.__name__
        if indices not in coeffs[fun_name]:
            coeffs[fun_name][indices] = function(*indices)

        return coeffs[fun_name][indices]

    res = defaultdict(lambda: 0)
    if m >= 0:
        for j in range(0, m + 1, 2):
            for k in range(0, l - m + 1, 2):
                for p in range(0, k + 1, 2):
                    for q in range(0, p + 1, 2):
                        coeff = (
                            (-1) ** ((j + p) // 2)
                            * get(A, l, m)
                            * get(B, l, m, j, k)
                            * get(C, p, q, k)
                        )
                        x_order = m - j + p - q
                        y_order = j + q
                        z_order = 0
                        res[(x_order, y_order, z_order)] += coeff
            for k in range(1, l - m + 1, 2):
                for p in range(0, k, 2):
                    for q in range(0, p + 1, 2):
                        coeff = (
                            (-1) ** ((j + p) // 2)
                            * get(A, l, m)
                            * get(B, l, m, j, k)
                            * get(C, p, q, k - 1)
                        )
                        x_order = m - j + p - q
                        y_order = j + q
                        z_order = 1
                        res[(x_order, y_order, z_order)] += coeff
    else:
        for j in range(1, abs(m) + 1, 2):
            for k in range(0, l - abs(m) + 1, 2):
                for p in range(0, k + 1, 2):
                    for q in range(0, p + 1, 2):
                        coeff = (
                            (-1) ** ((j + p - 1) // 2)
                            * get(A, l, abs(m))
                            * get(B, l, abs(m), j, k)
                            * get(C, p, q, k)
                        )
                        x_order = abs(m) - j + p - q
                        y_order = j + q
                        z_order = 0
                        res[(x_order, y_order, z_order)] += coeff
            for k in range(1, l - abs(m) + 1, 2):
                for p in range(0, k, 2):
                    for q in range(0, p + 1, 2):
                        coeff = (
                            (-1) ** ((j + p - 1) // 2)
                            * get(A, l, abs(m))
                            * get(B, l, abs(m), j, k)
                            * get(C, p, q, k - 1)
                        )
                        x_order = abs(m) - j + p - q
                        y_order = j + q
                        z_order = 1
                        res[(x_order, y_order, z_order)] += coeff

    return res


def p_coeffs(n):
    l = mp.floor(mp.sqrt(n))
    m = n - l * l - l
    mu = int(l - m)
    nu = int(l + m)
    if nu % 2 == 0:
        i = mu // 2
        j = nu // 2
        k = 0
    else:
        i = (mu - 1) // 2
        j = (nu - 1) // 2
        k = 1
    return (i, j, k)


def _A1(l_max):
    n = (l_max + 1) ** 2
    p = {m: p_coeffs(m) for m in range(n)}
    res = mp.zeros(n, n)

    k = 0
    for l in range(l_max + 1):
        for m in range(-l, l + 1):
            y = Y(l, m)
            res[:, k] = mp.matrix([y[p[i]] for i in range(n)])
            k += 1

    res = res * 2 / mp.sqrt(mp.pi)
    return res


def A1(l_max, cache=None):
    if cache is None:
        cache = CACHED_MATRICES
    return get_A(_A1, l_max, cache=cache)


def gtilde(n):
    l = mp.floor(mp.sqrt(n))
    m = n - l * l - l
    mu = l - m
    nu = l + m
    if nu % 2 == 0:
        I = [mp.floor(mu / 2)]
        J = [mp.floor(nu / 2)]
        K = [0]
        C = [mp.floor((mu + 2) / 2)]
    elif (l == 1) and (m == 0):
        I = [0]
        J = [0]
        K = [1]
        C = [1]
    elif (mu == 1) and (l % 2 == 0):
        I = [l - 2]
        J = [1]
        K = [1]
        C = [3]
    elif mu == 1:
        I = [l - 3, l - 1, l - 3]
        J = [0, 0, 2]
        K = [1, 1, 1]
        C = [-1, 1, 4]
    else:
        I = [mp.floor((mu - 5) / 2), mp.floor((mu - 5) / 2), mp.floor((mu - 1) / 2)]
        J = [mp.floor((nu - 1) / 2), mp.floor((nu + 3) / 2), mp.floor((nu - 1) / 2)]
        K = [1, 1, 1]
        C = [mp.floor((mu - 3) / 2), -mp.floor((mu - 3) / 2), -mp.floor((mu + 3) / 2)]
    res = {}
    for i, j, k, c in zip(I, J, K, C, strict=False):
        res[(i, j, k)] = c
    return res


def A2_inv(l_max):
    n = (l_max + 1) ** 2
    p = {m: p_coeffs(m) for m in range(n)}
    res = mp.zeros(n, n)

    k = 0
    for l in range(l_max + 1):
        for _ in range(-l, l + 1):
            y = gtilde(k)
            res[:, k] = mp.matrix([y[p[i]] if p[i] in y else 0.0 for i in range(n)])
            k += 1

    return res


def _A2(lmax):
    """Compute the A2 matrix directly

    A2_inv is way faster to compute but then computing the inverse with mpmath is very
    slow (~4 min for lmax=20). This function makes the computation by solving the change
    basis matrix for each column of the A2_inv matrix.

    The way it works is to formalize the problem by building the A matrix which
    encapsulate the linear system. The tricky part is to find all possible powers to
    use, which is done recursively in the all_subsets_indices function.
    """
    n = (lmax + 1) ** 2
    A2 = mp.zeros(n, n)
    gs = [basis.gtilde(i) for i in range(n)]

    def all_subsets_indices(s, sets, indices=None):
        if indices is None:
            indices = []
        for si in s:
            for i, set_ in enumerate(sets):
                if si in set_ and i not in indices:
                    indices.append(i)
                    all_subsets_indices(set_, sets, indices)
        return indices

    for k in range(n):
        available_sets = gs[0 : int((np.floor(np.sqrt(k)) + 1) ** 2)]
        target_set = [basis.ptilde(k)]
        indices = np.array(all_subsets_indices(target_set, available_sets))

        powers = list(reduce(set.union, [set(gs[i].keys()) for i in indices]))

        A = mp.zeros(len(powers), len(indices))

        for i, ijk in enumerate(powers):
            for j, index in enumerate(indices):
                A[i, j] = gs[index].get(ijk, 0)

        b = np.array([target_set[0] == p for p in powers]).astype(int)

        # in multi-precision
        x = mp.lu_solve(A, utils.to_mp(b))

        for i, _x in zip(indices, x, strict=False):
            A2[k, i] += _x

    return A2.T


def A2(lmax, cache=None):
    if cache is None:
        cache = CACHED_MATRICES
    return get_A(_A2, lmax, cache=cache)
